//
// Sound data classes for OpenAL
//

#ifdef	MACOSX
	#include	<OpenAL/al.h>
#else
	#include	<al.h>
#endif

#include	<stdio.h>
#include	<string.h>
#include	<vorbis/vorbisfile.h>
#include	<mad.h>
#include	"SoundData.h"
#include	"Data.h"
#include	"OggSoundData.h"
#include	"WavSoundData.h"
#include	"Mp3SoundData.h"

void	copyFirstChannel ( byte * dst, byte * src, int size, int numChannels, int bytesPerSample );

SoundData :: SoundData ( Data * theData )
{
	data          = theData;
	fp            = 0;
	numChannels   = 0;
	frequency     = 0;
	bitsPerSample = 8;
	ok            = true;
}

SoundData :: SoundData ( const char * theFileName )
{
	data          = NULL;
	fp            = fopen ( theFileName, "rb" );
	numChannels   = 0;
	frequency     = 0;
	bitsPerSample = 8;
	ok            = true;
}

SoundData :: ~SoundData ()
{
	delete data;

	if ( fp != NULL )
		fclose ( fp );
}

bool	SoundData :: isOk () const
{
	return data != NULL || fp != NULL && ok;
}

bool	SoundData :: atEnd () const
{
	if ( data != NULL )
		return data -> getPos () >= data -> getLength ();

	return true;
}

////////////////////// Concrete classes for wav/ogg/mp3 formats ////////////////
////////////////////////////////////// ogg  ////////////////////////////////////

OggSoundData :: OggSoundData  ( Data * theData ) : SoundData ( theData )
{
	ov_callbacks	cb;

	cb.read_func  = readFunc;
	cb.seek_func  = seekFunc;
	cb.close_func = closeFunc;
	cb.tell_func  = tellFunc;

	if ( ov_open_callbacks ( this, &file, NULL, 0, cb ) < 0 )
	{
		ok = false;

		return;
	}

	info          = ov_info ( &file, -1 );
	numChannels   = info -> channels;
	frequency     = info -> rate;
	bitsPerSample = 16;
}

OggSoundData :: OggSoundData  ( const char * theFileName ) : SoundData ( theFileName )
{
	if ( ov_open ( fp, &file, NULL, 0 )	< 0 )
	{
		ok = false;

		return;
	}

	info          = ov_info ( &file, -1 );
	numChannels   = info -> channels;
	frequency     = info -> rate;
	bitsPerSample = 16;
}

OggSoundData :: ~OggSoundData ()
{
	ov_clear ( &file );
}

size_t	OggSoundData :: getSize () const
{
	double	totalTime = ov_time_total ( const_cast <OggVorbis_File *> ( &file ), -1 );

	return (size_t)(totalTime * numChannels * frequency * 2);
}
											// fill buffer with next audio data
int		OggSoundData :: read ( byte * buf, int size )
{
	if ( size < 0 )
		size = getSize ();

	int	curSection, res;
	int	bytesRead = 0;

	while ( bytesRead < size )
	{
		res = ov_read ( &file, (char *)(buf + bytesRead), size - bytesRead, 0, 2, 1, &curSection );

		if ( res <= 0 )
			break;

		bytesRead += res;
	}

	return bytesRead;
}

int		OggSoundData :: readMono ( byte * buf, int size )
{
	if ( numChannels == 1 )
		return read ( buf, size );

	if ( size < 0 )
		size = getSize ();

	int		curSection, res;
	int		bytesRead = 0;
	byte	tmpBuf [10240];				// use 10K work buffer

	while ( bytesRead < size )
	{
		int	sz = size - bytesRead;

		if ( sz > sizeof ( tmpBuf ) )
			sz = sizeof ( tmpBuf );

		res = ov_read ( &file, (char *)tmpBuf, size - bytesRead, 0, 2, 1, &curSection );

		if ( res <= 0 )
			break;

		copyFirstChannel ( buf + bytesRead, tmpBuf, res, numChannels, bitsPerSample / 8 );

		bytesRead += res / numChannels;
	}

	return bytesRead;
}

											// position in sound (in seconds)
bool	OggSoundData :: seek ( float time )
{
	ov_time_seek ( &file, time );

	return true;
}

////////////////////////////////////// wav  ////////////////////////////////////

#pragma pack (push, 1)

struct WavFmt
{
		unsigned short encoding;
		unsigned short channels;
		unsigned int   frequency;
		unsigned int   byterate;
		unsigned short blockAlign;
		unsigned short bitsPerSample;
};

#pragma pack (pop)

WavSoundData :: WavSoundData ( Data * theData ) : SoundData ( theData )
{
	long	magic  = data -> getLong ();
	long	length = data -> getLong ();
	long	magic2 = data -> getLong ();

	if ( magic != RIFF || magic2 != WAVE )
	{
		ok = false;

		return;
	}

	for ( ; data -> getPos () < data -> getLength (); )
	{
		magic  = data -> getLong ();
		length = data -> getLong ();

		if ( magic == FMT )
		{
			WavFmt	format;

			data -> getBytes ( &format, sizeof ( format ) );

			if ( format.encoding != 1 )
			{
				ok = false;

				return;
			}

			numChannels   = format.channels;
			frequency     = format.frequency;
			bitsPerSample = format.bitsPerSample;
		}
		else
		if ( magic == DATA )
		{
			dataOffs   = data -> getPos ();
			dataLength = length;

			break;
		}
		else
			data -> seekCur ( length );
	}

	if ( numChannels < 1 || frequency == 0 || dataOffs == 0 || dataLength == 0 )
		ok = false;
}

WavSoundData :: ~WavSoundData ()
{

}

size_t	WavSoundData :: getSize () const
{
	return dataLength;
}

											// fill buffer with next audio data
int		WavSoundData :: read ( byte * buf, int size )
{
	int	bytesLeft = dataLength - (data -> getPos () - dataOffs);

	if ( size < 0 || size > bytesLeft )
		size = bytesLeft;

	return data -> getBytes ( buf, size );
}

int		WavSoundData :: readMono ( byte * buf, int size )
{
	if ( numChannels == 1 )
		return read ( buf, size );

	int	bytesLeft = dataLength - (data -> getPos () - dataOffs);

	if ( size < 0 || size > bytesLeft )
		size = bytesLeft;

	int		bytesRead = 0;
	byte	tmpBuf [10240];

	while ( bytesRead < size )
	{
		int	sz  = size - bytesRead;

		if ( sz > sizeof ( tmpBuf ) )
			sz = sizeof ( tmpBuf );

		int	res = data -> getBytes ( tmpBuf, size - bytesRead );

		if ( res <= 0 )
			break;

		copyFirstChannel ( buf + bytesRead, tmpBuf, res, numChannels, bitsPerSample / 8 );

		bytesRead += res / numChannels;
	}

	return bytesRead;

}

											// position in sound (in seconds)
bool	WavSoundData :: seek ( float time )
{
	data -> seekAbs ( dataOffs + (int)(time * numChannels * frequency*bitsPerSample/2) );	// XXX - 16 bit ???

	return true;
}

////////////////////////////////////// mp3 ////////////////////////////////////

Mp3SoundData :: Mp3SoundData ( const char * theFileName ) : SoundData ( theFileName )
{
	if ( fp == NULL )
	{
		ok = false;

		return;
	}

	fseek ( fp, 0, SEEK_END );

	fileSize = ftell ( fp );

	fseek ( fp, 0, SEEK_SET );

	bufferLength = 0;

	mad_stream_init ( &stream );
	mad_frame_init  ( &frame  );
	mad_synth_init  ( &synth  );
	mad_timer_reset ( &timer  );

	if ( !readFrame () )
	{
		ok = false;

		return;
	}

	numChannels   = ( frame.header.mode == MAD_MODE_SINGLE_CHANNEL ? 1 : 2 );
	frequency     = frame.header.samplerate;
	bitRate       = frame.header.bitrate;
	bitsPerSample = 16;
}

Mp3SoundData :: Mp3SoundData ( Data * theData ) : SoundData ( theData )
{
	fileSize     = data -> getLength ();
	bufferLength = 0;

	data -> seekAbs ( 0 );

	mad_stream_init ( &stream );
	mad_frame_init  ( &frame  );
	mad_synth_init  ( &synth  );
	mad_timer_reset ( &timer  );

	if ( !readFrame () )
	{
		ok = false;

		return;
	}

	numChannels   = ( frame.header.mode == MAD_MODE_SINGLE_CHANNEL ? 1 : 2 );
	frequency     = frame.header.samplerate;
	bitRate       = frame.header.bitrate;
	bitsPerSample = 16;
}

Mp3SoundData :: ~Mp3SoundData ()
{
	mad_synth_finish  ( &synth  );
	mad_stream_finish ( &stream );
	mad_frame_finish  ( &frame  );
}

bool	Mp3SoundData :: readFrame ()
{
	while ( true )
	{
		if ( stream.buffer == NULL || stream.error == MAD_ERROR_BUFLEN )
		{
			int		remaining;
			int		readSize;
			byte  * readStart;

			if ( stream.next_frame != NULL )
			{
				remaining = stream.bufend - stream.next_frame;
				readStart = buffer + remaining;
				readSize  = BUFFER_SIZE - remaining;

				memmove ( buffer, stream.next_frame, remaining );
			}
			else
			{
				readSize  = BUFFER_SIZE;
				readStart = buffer;
				remaining = 0;
			}

												// read more data
			if ( fp != NULL )
				readSize = fread ( readStart, 1, readSize, fp );
			else
				readSize = data -> getBytes ( readStart, readSize );

			if ( readSize < 0 )					// we're at end (check !!!)
				return 1;

												// check for EOF, use MAD_BUFFER_GUARD
			if ( (fp != NULL && feof ( fp ) ) || ( data -> getPos () >= data -> getLength () ) )
			{
				memset ( readStart + readSize, '\0', MAD_BUFFER_GUARD );

				readSize += MAD_BUFFER_GUARD;
	        }

			mad_stream_buffer ( &stream, buffer, readSize + remaining );

			stream.error = static_cast<enum mad_error> ( 0 );
		}


		if ( mad_frame_decode ( &frame, &stream ) )
		{
			if ( MAD_RECOVERABLE ( stream.error ) )
				continue;
			else
			if ( stream.error == MAD_ERROR_BUFLEN )
				continue;
			else
				break;	
		}
		else
			return true;
}

	return false;
}

size_t 	Mp3SoundData :: getSize () const
{
	return (size_t)((16 * (uint64)fileSize * (uint16) numChannels * (uint16) frequency) / bitRate);
}

bool	Mp3SoundData :: seek ( float time )
{
	unsigned	pos = (unsigned)((double)bitRate / 8.0 * time);

	if ( fp != NULL )
		fseek ( fp, pos, SEEK_SET );
	else
		data -> seekAbs ( pos );

	readFrame ();

	return true;
}

inline	int	resample ( mad_fixed_t sample )
{
	sample += 1 << (MAD_F_FRACBITS - 16);

	if ( sample >= MAD_F_ONE )
		sample = MAD_F_ONE - 1;
	else
	if ( sample < -MAD_F_ONE )
		sample = -MAD_F_ONE;

	return (int)(sample >> (MAD_F_FRACBITS  - 15));
}

int 	Mp3SoundData :: read ( byte * out, int size )
{
	if ( size < 0 )
		size = getSize ();

	int bytesRead = 0;

	while ( bytesRead < size )
	{
		mad_synth_frame ( &synth, &frame );

		mad_pcm     * pcm   = &synth.pcm;
		mad_fixed_t * left  = pcm -> samples [0];
		mad_fixed_t * right = pcm -> samples [1];
		uint16      * data  = (uint16 *)(out + bytesRead);

		for ( size_t length = pcm -> length; length > 0; length-- )
		{
			*data++ = (uint16) resample ( *left++ );

			if ( numChannels == 2 )
				*data++ = (uint16) resample ( *right++ );
		}

		bytesRead += pcm -> length * numChannels * 2;

		if ( !readFrame ()  )
			return bytesRead;
	}

	return bytesRead;
}

int 	Mp3SoundData :: readMono ( byte * out, int size )
{
	if ( size < 0 )
		size = getSize ();

	int bytesRead = 0;

	while ( bytesRead < size )
	{
		mad_synth_frame ( &synth, &frame );

		mad_pcm     * pcm   = &synth.pcm;
		mad_fixed_t * left  = pcm -> samples [0];
		uint16      * data  = (uint16 *)(out + bytesRead);

		for ( size_t length = pcm -> length; length > 0; length-- )
			*data++ = (uint16) resample ( *left++ );


		bytesRead += pcm -> length * 2;

		if ( !readFrame ()  )
			return bytesRead;
	}

	return bytesRead;
}

void	copyFirstChannel ( byte * dst, byte * src, int size, int numChannels, int bytesPerSample )
{
	int	samples = size / (bytesPerSample * numChannels);

	if ( bytesPerSample == 1 )
	{
		while ( --samples > 0 )
		{
			*dst++ = *src;
			src   += numChannels;
		}
	}
	else
	if ( bytesPerSample == 2 )
	{
		uint16 * d = (uint16 *) dst;
		uint16 * s = (uint16 *) src;

		while ( --samples > 0 )
		{
			*d++ = *s;
			s   += numChannels;
		}
	}
}
