//
// Sound source wrapper for OpenAL
//
// Author: Alex V. Boreskoff
//

#include	<stdio.h>
#include	"SoundSource.h"
#include	"SoundData.h"
#include	"OpenAlContext.h"

SoundSource :: SoundSource  ( SoundData * data, bool theLoop, bool forceMono )
{
	dataSource = data;

	if ( dataSource == NULL )
		return;

	looping = theLoop;
	stream  = true;
	format  = soundFormat ( data -> getNumChannels (), data -> getBitsPerSample () );
	mono    = forceMono;

	if ( mono )
		format = soundFormat ( 1, data -> getBitsPerSample () );

	if ( dataSource -> getSize () < BUFFER_SIZE * 2 )
		stream = false;

	alGenSources ( 1, &source );

	if ( stream )
	{
		buf           = new byte [BUFFER_SIZE * 2];
		currentBuffer = 0;

		alGenBuffers ( 2, buffer  );
	}
	else
	{
		buf = new byte [BUFFER_SIZE * 2];

		int size = mono ? dataSource -> readMono ( buf ) : dataSource -> read ( buf );

		alGenBuffers ( 1, buffer );
		alBufferData ( buffer[0], format, buf, size, dataSource -> getFrequency () );

		alSourcei ( source, AL_BUFFER, buffer[0] );
		alSourcei ( source, AL_LOOPING, looping ? AL_TRUE : AL_FALSE );

		delete buf;
	}
}

SoundSource :: ~SoundSource ()
{
	if( dataSource == NULL )
		return;

	delete dataSource;

	if ( stream )
		delete buf;

	alDeleteSources ( 1, &source );

	if ( stream )
		alDeleteBuffers ( 2, buffer );
	else
		alDeleteBuffers ( 1, buffer );
}

ALuint	SoundSource :: soundFormat ( int numChannels, int bitsPerSample ) const
{
	if ( numChannels == 1 )
		return bitsPerSample == 16 ? AL_FORMAT_MONO16 : AL_FORMAT_MONO8;

	if ( numChannels == 2 )
		return bitsPerSample == 16 ? AL_FORMAT_STEREO16 : AL_FORMAT_STEREO8;
		
	if ( numChannels == 4 )
		if ( bitsPerSample == 16 )
			return alGetEnumValue ( "AL_FORMAT_QUAD16" );
		else
			return alGetEnumValue ( "AL_FORMAT_QUAD8" );
			
	if ( numChannels == 6 )
		if ( bitsPerSample == 16 )
			return alGetEnumValue ( "AL_FORMAT_51CHN16" );
		else
			return alGetEnumValue ( "AL_FORMAT_51CHN8" );
			
	return 0;
}

bool	SoundSource :: isPlaying () const
{
	ALint state;

	alGetSourcei ( source, AL_SOURCE_STATE, &state );

	return state == AL_PLAYING;
}

bool	SoundSource :: atEnd () const
{
	if ( isPlaying () )
		return false;

	return dataSource -> atEnd ();
}

void	SoundSource :: setPos ( const Vector3D& newPos )
{
	pos = newPos;

	alSourcefv ( source, AL_POSITION, pos );
}

void	SoundSource :: setVelocity ( const Vector3D& newVel )
{
	velocity = newVel;

	alSourcefv ( source, AL_VELOCITY, velocity );
}

void	SoundSource :: setPitch ( float newPitch )
{
	pitch = newPitch;

	alSourcef ( source, AL_PITCH, pitch );
}

void	SoundSource :: setGain ( float newGain )
{
	gain = newGain;

	alSourcef ( source, AL_GAIN, gain );
}

void	SoundSource :: setLooping ( bool flag )
{
	looping = flag;

	alSourcei ( source,AL_LOOPING, looping ? AL_TRUE : AL_FALSE );
}

void	SoundSource :: play ()
{
	ALint 	state;
	int		size;

	alGetSourcei ( source, AL_SOURCE_STATE, &state );

	if ( state == AL_PLAYING )
		return;

	if ( state != AL_PAUSED && stream )
	{
		if ( mono )
			size = dataSource -> readMono ( buf, BUFFER_SIZE );
		else
		    size = dataSource -> read ( buf, BUFFER_SIZE );

		alBufferData ( buffer [0], format, buf, size, dataSource -> getFrequency () );

		if ( mono )
			size = dataSource -> readMono ( buf, BUFFER_SIZE );
		else
		    size = dataSource -> read ( buf, BUFFER_SIZE );

		alBufferData         ( buffer [1], format, buf, size, dataSource -> getFrequency () );
		alSourceQueueBuffers ( source, 2, buffer );

		OpenAlContext :: registerSound ( this );
	}

	alSourcePlay ( source );
}

void	SoundSource :: pause ()
{
	alSourcePause ( source );
}

void	SoundSource :: rewind ()
{
	alSourceRewind ( source );
}

void	SoundSource :: stop ()
{
	alSourceStop ( source );

	dataSource -> seek ( 0 );

	if ( stream )
	{
		ALint queued;

		alGetSourcei ( source,AL_BUFFERS_QUEUED,&queued );

		if ( queued > 0 )
			alSourceUnqueueBuffers ( source, 2, buffer );

		OpenAlContext :: unregisterSound ( this );

		currentBuffer = 0;
	}
}

void	SoundSource :: update ()
{
	if ( stream )
	{
		ALint processed;

		alGetSourcei ( source, AL_BUFFERS_PROCESSED, &processed );

		if ( processed == 0 )
			return;

		if ( processed == 1 )
		{
			alSourceUnqueueBuffers ( source, 1, &buffer [currentBuffer] );

			int size = mono ? dataSource -> readMono ( buf, BUFFER_SIZE ) : dataSource -> read ( buf, BUFFER_SIZE );

			if ( size > 0 || (size == 0 && looping) )
			{
				alBufferData         ( buffer [currentBuffer], format, buf, size,
				                       dataSource -> getFrequency () );

				alSourceQueueBuffers ( source, 1, &buffer [currentBuffer] );

				if ( size < BUFFER_SIZE && looping )
					dataSource -> seek ( 0 );
			}
			else
			{
				int queued;

				alGetSourcei ( source, AL_BUFFERS_QUEUED, &queued );

				if ( queued == 0 )
					dataSource -> seek ( 0 );
			}

			currentBuffer = 1 - currentBuffer;
		}
		else
		if ( processed == 2 )
		{
			alSourceUnqueueBuffers ( source, 2, buffer );

			currentBuffer = 0;

			play ();
		}
	}
}
